function [yhat,h,m,ybar,variance]=addfit(x,y,p)

% This function computes the additive model fit for 
%
%             y = a + m_1(x_1) + ... + m_D(x_D) + iid errors
%
% using local polynomials of degree p=p_1,...,p_D.
% 
% Syntax: [yhat,h,m,ybar,variance]=addfit(x,y,p)
%
% Inputs:  x   matrix of covariates (nxD)
%          y   vector of observations (nx1)
%          p   order of local polynomials for additive functions (1xD or scalar)
%
% Outputs: yhat      vector fits at the observation points (nx1)
%          h         bandwidths for the D additive functions (1xD)
%          m         matrix of fits for the additive functions at the
%                    observation points (nxD)
%          ybar      mean of y and estimate of a (scalar)
%          variance  estimated variance of the errors (scalar)
%
% NOTE: THE ORDER p OF THE LOCAL POLYNOMIALS HAS TO BE ODD!
%
% Calls the following m-files: backfit, lpolyregh, polyreg, fact, constants
%
% This additive model fitting program implements the backfitting 
% algorithm and bandwidth selection method described in Opsomer and Ruppert
% (Journal of the American Statistical Association, forthcoming).
%
%	                         Last modified: 8/8/97 by JDO

% Initialization (sets the meta-parameters)

%     numbin is the number of bins into which the observations are distributed
%     for each covariate, which is used in the computation of a minimum "safe"
%     bandwidth to avoid singularity
%                                           (should normally not be adjusted)
%     marg is the inflation factor used to increase the minimum "safe" 
%     bandwidth
%                                   (can be increased in case of singularity)

numbin=500;
marg=0.2;

%    Cmax is the maximum number of iterations in the search for the optimum
%    number of "pieces" for the piecewise polynomial regression
%                                           (should normally not be adjusted)
%    Nmax is the maximum number of pieces for each covariate
%                                   (can be decreased in case of singularity)
Cmax=10;
Nmax=7;

%    alpha is the proportion of each boundary that is ignored in the
%    computation of theta11
%                               (can be increased if plot of theta11 displays
%                                              instability at the boundaries)

alpha=0.1;

%    M1 is the number of gridpoints in the backfitting algorithm
%                                         (can be decreased to improve speed)

M1=100;

% Compute the constants for the bandwidth estimators for local polynomial
% degree p

cst=constants(p);

% Compute the minimum "safe" bandwidths for theta11 (bmg), variance (bmh)
% and m (bmh)

[n,D]=size(x);
if n<=D
	'*** matrix x has more covariates than observations ***'
	'*** possible transpose error? ***'
	stop
end
numbin=n*(n<=numbin)+numbin*(n>numbin);
Nmax= max(min(fix(n/(D*20)),Nmax),1);
ybar=sum(y)/n;
ycen=y-ybar;
rx=max(x)-min(x);

bmg=zeros(1,D);
bmh=bmg;
if max(size(p))==1
	p=p*ones(1,D);
	cst=cst*ones(1,D);
end

for i=1:D
	[nn,xx]=hist(x(:,i),numbin);
	xx=xx(find(nn~=0))';
	nn=max(size(xx));
	temp=sort(xx(1:nn-p(i)-3)-xx(p(i)+4:nn));
	bmg(i)=-temp(1)*(1+marg);
	temp=sort(xx(1:nn-p(i)-1)-xx(p(i)+2:nn));
	bmh(i)=-temp(1)*(1+marg);
end

% Find initial values of sigma2 and theta13 by piecewise polynomial 
% regression of degree p+4

pp=[];
xx=[];
for i=1:D
	for j=1:Nmax
		tempo=zeros(n,1);
		eval(['xx',int2str(i),int2str(j),'=[];'])
		eval(['pp',int2str(i),int2str(j),'=[];'])
		for jj=1:j
			temp=(x(:,i) <= min(x(:,i))+jj/j*rx(i));
			eval(['xx',int2str(i),int2str(j),'=[xx',int2str(i),int2str(j),' x(:,i).*(temp-tempo)];'])
			tempo=temp;
			eval(['pp',int2str(i),int2str(j),'=[pp',int2str(i),int2str(j),' p(i)];'])
		end
	end
	eval(['xx=[xx xx',int2str(i),int2str(Nmax),'];'])
	eval(['pp=[pp pp',int2str(i),int2str(Nmax),'];'])
end

yhat=polyreg(y,xx,pp+4);
rsmax=sum((yhat-y).^2);
dfmax=(sum(p)+4*D)*Nmax;
rs=zeros(1,Nmax);
minvec=zeros(1,D);
minvecnew=ones(1,D);
count=0;
xx=x;
pp=p;

while (any(minvecnew~=minvec) & count < Cmax)
  xxs=[];
  pps=[];
  yhat=polyreg(y,xx,pp+4);
  rsm=sum((yhat-y).^2);
  minvec=minvecnew;
  minvecs=[0 cumsum(minvec)];
  for i=1:D
    for j=1:Nmax
      if j==minvec(i)
        rs(j)=rsm;
      else
        minl=minvecs(i);
        maxl=minvecs(i+1)+1;
        tot=minvecs(i+1);
        eval(['xxt=[xx(:,1:minl) xx',int2str(i),int2str(j),' xx(:,maxl:tot)];'])
        eval(['ppt=[pp(:,1:minl) pp',int2str(i),int2str(j),' pp(:,maxl:tot)];'])
        yhat=polyreg(y,xxt,ppt+4);
        rs(j)=sum((yhat-y).^2);
      end
    end
    df=(p(i)+4)*linspace(1,Nmax,Nmax)+minvec*(p'+4)-minvec(i)*(p(i)+4);
    Cp=rs/n+2*(1+df)/n*rsmax/(n-1-dfmax);
    [Cpmin,jmin]=min(Cp);
    minvecnew(i)=jmin;
    eval(['xxs=[xxs xx',int2str(i),int2str(jmin),'];'])
    eval(['pps=[pps pp',int2str(i),int2str(jmin),'];'])
  end
  xx=xxs;
  pp=pps;
  count=count+1;
end

if count==Cmax
'*** Cp count exceeded ***'
end

[yhat,beta]=polyreg(y,xx,pp+4);
df=max(size(beta))-1;
sigma2P=sum((y-yhat).^2)/(n-1-df);

mstart=zeros(size(x));
theta1=zeros(size(x));
theta3=zeros(size(x));
beta(1)=[];
for i=1:D
	for j=1:minvecnew(i)
		btemp1=beta(p(i)+1:p(i)+4);
		p1=fact(linspace(p(i)+1,p(i)+4,4)')./fact([0 1 2 3]');
		btemp1=btemp1.*p1;
		btemp3=beta(p(i)+3:p(i)+4);
		p3=fact([p(i)+3 p(i)+4]');
		btemp3=btemp3.*p3;
		Dp=[(xx(:,1)~=0) xx(:,1)];
		theta3(:,i)=theta3(:,i)+Dp*btemp3;
		Dp=[Dp xx(:,1).^2 xx(:,1).^3];
		theta1(:,i)=theta1(:,i)+Dp*btemp1;
		binit=[0; beta(1:p(i)+4)];
		for k=4:p(i)+4
			Dp=[Dp xx(:,1).^k];
		end
		mstart(:,i)=mstart(:,i)+Dp*binit;
		beta(1:p(i)+4)=[];
		xx(:,1)=[];
	end
end

theta1=theta1-ones(n,1)*sum(theta1)/n;
theta3=theta3-ones(n,1)*sum(theta3)/n;
mstart=mstart-ones(n,1)*sum(mstart)/n;
theta13=sum(theta1.*theta3)/n;

% Compute theta11 through additive model with local polynomial regression
% of degree p+2

g=cst(2,:)+(cst(3,:)-cst(2,:)).*(theta13>0);
g=g.*(sigma2P*rx./abs(theta13)/n).^(1./(2*p+5));
for i=1:D
	if g(i)<bmg(i)
		g(i)=bmg(i);
	end
end

thet=zeros(n,D);
clip=thet;
xc=(min(x)+max(x))/2;
minlim=min(x)+alpha*rx;
maxlim=max(x)-alpha*rx;
m=backfit(ycen,x,g,p+2,mstart,M1);
if(n>M1)
	grid=linspace(0,1,M1);
	xgrid=ones(M1,1)*(xc-1.01*rx/2)+grid'*(xc+1.01*rx/2);
	for i=1:D
		if(D==1)
			temp=lpolyregh(xgrid(:,i),x(:,i),g(i),p(i)+2,p(i)+1)*ycen;
		else
			temp=lpolyregh(xgrid(:,i),x(:,i),g(i),p(i)+2,p(i)+1)*(ycen-sum(m')'+m(:,i));
		end
		thet(:,i)=interp1(xgrid(:,i),temp,x(:,i),'cubic');
		clip(:,i)=(minlim(i)<x(:,i)).*(x(:,i)<maxlim(i));
	end
else
	xgrid=x;
	if(D==1)
		thet(:)=lpolyregh(xgrid(:),x(:),g,p+2,p+1)*ycen;
	else
		for i=1:D
			thet(:,i)=lpolyregh(xgrid(:,i),x(:,i),g(i),p(i)+2,p(i)+1)*(ycen-sum(m')'+m(:,i));
			clip(:,i)=(minlim(i)<x(:,i)).*(x(:,i)<maxlim(i));
		end
	end
end
thet=thet.*clip;
means=zeros(1,D);
theta11=zeros(1,D);
for i=1:D
	t=thet(~isnan(thet(:,i)),i);
	clipn=clip(~isnan(thet(:,i)),i);
	means(i)=mean(t);
	t=t-means(i);
	t=t.*clipn;
	theta11(i)=sum(t.^2)./max(size(t));
end

% Compute variance
k=cst(4,:).*(sigma2P*rx./theta11/n).^(1./(2*p+3));
for i=1:D
	if k(i)<bmh(i)
		k(i)=bmh(i);
	end
end

m=backfit(ycen,x,k,p,mstart,M1);
if(D==1)
	yhat=ybar+m;
else
	yhat=ybar+sum(m')';
end
variance=sum((y-yhat).^2)/n;

% Compute additive fit with local polynomials of degree p

h=cst(1,:).*(variance*rx./theta11/n).^(1./(2*p+3));
for i=1:D
	if h(i)<bmh(i)
		h(i)=bmh(i);
	end
end

[m]=backfit(ycen,x,h,p,mstart,M1);
if(D==1)
	yhat=ybar+m;
else
	yhat=ybar+sum(m')';
end
